


# using Polynomials,
# using LightGraphs
# using Random

# getRNG(seed::Integer=-1) = seed >= 0 ? MersenneTwister(seed) : GLOBAL_RNG

function erdos_renyi_nm(n::Integer, m::Integer; tries=3, seed=-1, keyargs...)
	G = try_until_connected(erdos_renyi, tries, (n, m), seed; keyargs...)
	return G
end

gen_er = erdos_renyi_nm

"""
	barabasi_albert_nm(n::Integer, m::Integer; tries=3, seed=-1, keyargs...)
Can also call this using ```gen_sf``` with the same arguments. Generate a Barabási–Albert random graph with ```n``` nodes and as close to ```m``` edges as possible (note that due to how such graphs are constructed, certain target densities are difficult to achieve). Throws an error when it is not mathematically possible.

### Optional arguments
- ```tries=3```: number of regenerations to try if the graph is not connected. Returns if the graph is connected or if all tries are used.
- ```keyargs...```: see [LightGraphs' ```barabasi_albert```](https://juliagraphs.org/LightGraphs.jl/stable/generators/#LightGraphs.SimpleGraphs.barabasi_albert-Tuple{Integer,Integer})

### Implementation notes
```m = nk - k^2```, where ```k``` is the parameter from LightGraphs' ```barabasi_albert``` method. Thus, we use the quadratic formula to solve for ```k```. However, since ```k``` must be an integer to be inputted into ```barabasi_albert```, rounding is needed and thus achieving an exact match with ```m``` is not always possible. It is possible to first generate a smaller graph and build upon it, but such an implementation is a bit more complicated and left for a later date.
"""
function barabasi_albert_nm(n::Integer, m::Integer; tries=3, seed=-1, keyargs...)
	r1 = (-n + (n^2 - 4 * m)^0.5) / -2
	r2 = (-n - (n^2 - 4 * m)^0.5) / -2
	kroots = [r1, r2]
	# kpoly = Polynomial([0, n, -1]) - m
	# kroots = roots(kpoly)
	realroots = filter(isreal, kroots)
	length(realroots) == 0 && throw(DomainError(m, "not possible to create a barabasi_albert graph with $n nodes and $m edges"))
	validroots = map(x -> convert(Float64, x), realroots)
	k = convert(Int, round(realroots[1]))
	G = try_until_connected(barabasi_albert, tries, (n, k), seed; keyargs...)
	return G
	# return barabasi_albert(n, k; keyargs...)
end

gen_sf = barabasi_albert_nm


function euclidean_graph_nodist(args...; kwargs...)
	G, dists = euclidean_graph(args...; kwargs...)
	return G
end

"""
	euclidean_graph_nm()
Can also call this using ```gen_geo``` with the same arguments. Generate a geometric random graph with ```n``` nodes and as close to ```m``` edges as possible. Throws an error when it is not mathematically possible.

### Optional arguments

### Implementation notes
Given ```n``` and ```m```, we can solve for the cutoff as follows:

For a given node, within the sphere of radius ```cutoff```, for a cube of length ```L```, on average, the fraction of other points in that sphere is ```((4/3) * π * cutoff^3) / (L^3)```. For the first node, that fraction times ```n-1``` edges (the maximum number of possible edges) are formed. For the second node, that fraction times ```n-2``` edges (since the edge going between the first and second node was already checked) are formed, etc. Thus, the expected number of edges is that fraction multiplied by ```n(n-1)/2```. Solve for ```cutoff```. Note that for this solution to work, the boundary condition must be ```:periodic```, distance must be Euclidean (```p=2```), and it must be in 3 dimensions (```d=3```), so they are no longer optional parameters.
"""
function euclidean_graph_nm2(n::Integer, m::Integer; tries=3, seed=-1, L=1.)
	d = 3
	cutoff = (3 * m * L^d / (2 * π * n * (n-1)))^(1/3)
	cutoff > L && throw(DomainError(m, "not possible to create a euclidean_graph graph with $n nodes and $m edges, likely because $m > ($n * ($n-1) / 2), which is the maximum number of possible edges."))
	partial_euclidean_graph(x, y; keyargs...) = euclidean_graph_nodist(x, y; cutoff=cutoff, L=L, bc=:periodic, p=2, keyargs...)
	G = try_until_connected(partial_euclidean_graph, tries, (n, d), seed)
	# euclidean_graph_nodist(n, 3; L=100, cutoff=cutoff, keyargs...)
	return G
end




# function euclidean_graph_topm(n, m; d=3, L=1., p=2., bc=:open, seed=-1)
# 	G, weights = euclidean_graph(n, d; L=L, seed=seed, p=p, bc=bc)
#
# 	sorted_dists = sort(collect(weights), by=x->x[2], rev=true)
# 	bot = sorted_dists[m+1:end]
# 	for tup in bot
# 		e = tup[1]
# 		source = e[1]
# 		dest = e[2]
# 		rem_edge!(G, source, dest)
# 		rem_vertex!(G, source)
# 		rem_vertex!(G, dest) # this complicates things since
# 	end
# end

function _check_and_update_map!(d, val, counter)
	outval = 0
	if val in keys(d)
		outval = d[val]
	else
		d[val] = counter
		outval = counter
		counter += 1
	end
	return outval, counter
end

function _map_to_sequential!(arr)
	num2sequential = Dict{Int, Int}()
	currentind = 1
	mapcounter = 1
	mx = 0
	for (tind, t) in enumerate(arr)
		source = src(t)
		dest = dst(t)
		newsource, mapcounter = _check_and_update_map!(num2sequential, source, mapcounter)
		newdest, mapcounter = _check_and_update_map!(num2sequential, dest, mapcounter)
		if newsource >= mx
			mx = newsource
		end
		if newdest >= mx
			mx = newdest
		end
		arr[tind] = Edge(newsource, newdest)
	end
	return mx

end

function euclidean_graph_top(n, m; d=3, L=1, p=2, bc=:open, seed=-1)
	m > binomial(n, 2) && throw(DomainError(m, "not possible to create a euclidean_graph graph with $n nodes and $m edges because $m > ($n * ($n-1) / 2), which is the maximum number of possible edges."))
	if seed > -1
		Random.seed!(seed)
	end
	points = rmul!(rand(d, n), L)
	edge_dist_pq = PriorityQueue{Edge{Int64}, Float64}()
	for i in 1:n
		for j in (i+1):n
			if bc == :open
				Δ = points[:, i] - points[:, j]
			elseif bc == :periodic
				Δ = abs.(points[:, i] - points[:, j])
				Δ = min.(L .- Δ, Δ)
			else
				throw(ArgumentError("$bc is not a valid boundary condition"))
			end
			dist = norm(Δ, p)
			enqueue!(edge_dist_pq, Edge(i, j), dist)
			# indextoinsert = _sorted_insert!(sorted_dists, dist)
			# insert!(sorted_edges, indextoinsert, Edge(i, j))
		end
	end

	edgearr = []
	for i in 1:m
		e = dequeue!(edge_dist_pq) # smallest distances
		push!(edgearr, e)
	end
	mx = _map_to_sequential!(edgearr)
	G = SimpleGraph(mx)
	for e in edgearr
		add_edge!(G, e)
	end
	if nv(G) < n
        add_vertices!(G, n - nv(G))
    end
	return G
end

function euclidean_graph_nm(n, m; d=3, L=1, p=2, bc=:open, seed=-1, tries=3)

	partial_euclidean_graph(x, y; keyargs...) = euclidean_graph_top(x, y; d=d, L=L, p=p, bc=bc, seed=seed)
	G = try_until_connected(partial_euclidean_graph, tries, (n, m), seed)
	# euclidean_graph_nodist(n, 3; L=100, cutoff=cutoff, keyargs...)
	return G
end

gen_geo = euclidean_graph_nm


function try_until_connected(model, tries, model_params, seed; kwargs...)
	G = model(model_params...; seed=seed, kwargs...)
	# rng = getRNG(seed)
	if seed > -1
		Random.seed!(seed)
	end
	# println(rng)
	trycounter = 0
	while true
		if trycounter >= tries || length(connected_components(G)) == 1
			break
		end
		# println(trycounter)
		randadd = rand(Int)
		G = model(model_params...; seed=seed+randadd, kwargs...)
		trycounter += 1
	end
	return G
end

function gen_synthetic_non(l1rgms, l1names, l1n, l1m, l2rgms, l2names, l2n, l2m, l2numgroupsper, l2acrosspercent; rewirepercent=0.0, numprocs=1, seed=0)

	if ("sf" in l1names && l1names[1] != "sf") || ("sf" in l2names && l2names[1] != "sf")
		throw(ArgumentError("Because SF is not guaranteed to reach target size, it needs to be listed as the first rgm"))
	end

	Random.seed!(seed)
	numl1rgms = length(l1rgms)
	numl2rgms = length(l2rgms)
	numlabs = numl1rgms * numl2rgms
	totalnumgroups = numlabs * l2numgroupsper

	l2macross = round(Int, l2acrosspercent * l2m)
	l2mper = round(Int, (l2m - l2macross) / totalnumgroups)
	l2nper = round(Int, l2n / totalnumgroups)

	l1rgms_repeated = repeat(l1rgms, inner=numl2rgms)
	l1names_repeated = repeat(l1names, inner=numl2rgms)
	l2rgms_repeated = repeat(l2rgms, outer=numl1rgms)
	l2names_repeated = repeat(l2names, outer=numl1rgms)

	println(l1names_repeated)
	# println(l1rgms_repeated)
	println(l2names_repeated)
	# println(l2rgms_repeated)

	l2glist = AbstractGraph[]
	l2tags = String[]
	l1glist = AbstractGraph[]
	l1tags = String[]
	l2tol1 = Dict()
	l1graphcounter = 1

	sf_sizes = []
	l1_sf_sizes = []
	for (mind, l2model) in enumerate(l2rgms_repeated)
		println("On level 2, model $mind of type $(l2names_repeated[mind])")


		for i in 1:l2numgroupsper

			Gind = (mind-1) * l2numgroupsper + i
			if "sf" in l2names && l2names_repeated[mind] != "sf"
				l2min = sf_sizes[Gind % length(sf_sizes)+1]
			else
				l2min = l2mper
			end
			# println(Gind)
			seedval = rand(1:1000000000)
			G = l2model(l2nper, l2min; seed=seedval)
			if l2names_repeated[mind] == "sf"
				push!(sf_sizes, length(edges(G)))
			end
			rewire_edges!(G, rewirepercent; seed=seedval)
			push!(l2glist, G)
			push!(l2tags, l2names_repeated[mind])
			tempglist = Array{AbstractGraph}(undef, nv(G))
			temptlist = Array{String}(undef, nv(G))

			for vind in 1:nv(G)
				v = vertices(G)[vind]
				l2tol1[(Gind, vind)] = l1graphcounter
				l1graphcounter += 1
			end
			println("\tOn level 1, model $mind of type $(l1names_repeated[mind])")

			for vind in 1:nv(G)
				if "sf" in l1names && l1names_repeated[mind] != "sf"
					l1min = l1_sf_sizes[l2tol1[(Gind, vind)]%length(l1_sf_sizes)+1]
				else
					l1min = l1m
				end
				l1model = l1rgms_repeated[mind]
				seedval = rand(1:1000000000)
				l1G = l1model(l1n, l1min; seed=seedval)
				if l1names_repeated[mind] == "sf"
					push!(l1_sf_sizes, length(edges(l1G)))
				end
				rewire_edges!(l1G, rewirepercent; seed=seedval)
				tempglist[vind] = l1G
				temptlist[vind] = l1names_repeated[mind]
			end
			append!(l1glist, tempglist)
			append!(l1tags, temptlist)

		end
	end

	l1GG = GraphOfGroups(l1glist, l1tags)
	l2GG = GraphOfGroups(l2glist, l2tags, [InterEdge((0, 0), (0, 0))])
	numedegestoadd = l2macross
	while numedegestoadd > 0
		G1ind = rand(1:length(l2GG.groups))
		G2ind = rand(1:length(l2GG.groups)) # randomly pick two groups
		if G1ind == G2ind
			continue
		end
		n1ind = rand(1:nv(l2GG.groups[G1ind]))
		n2ind = rand(1:nv(l2GG.groups[G2ind])) # randomly pick a node from each group
		# n1ind is allowed to equal n2ind since they are in different graphs
		# in the case of SimpleGraphs, n1ind == n1. So for simplicity, don't getindex the actual node
		ie = InterEdge((G1ind, n1ind), (G2ind, n2ind))
		if findfirst(isequal(ie), l2GG.interedges) == nothing
			add_interedge!(l2GG, ie)
			numedegestoadd -= 1
		end
	end

	l1 = NoN(l1GG)
	l2 = NoN(l2GG)
	assign_lower!(l2, l1, l2tol1)
	return l1
end

function name2function(mname)
    if mname == "sf"
        return gen_sf
    elseif mname == "geo"
        return gen_geo
    else
        return
    end
end

function gen_synthetic_non(d::Dict; numprocs=1, seed=0)
	# l1rgms, l1names, l1n, l1m, l2rgms, l2names, l2n, l2m, l2numgroupsper, l2acrosspercent; rewirepercent=0.0, numprocs=1, seed=0)
	l1rgms = map(name2function, d["l1models"])
	l2rgms = map(name2function, d["l2models"])
	return gen_synthetic_non(l1rgms, d["l1models"], d["l1n"], d["l1m"], l2rgms, d["l2models"], d["l2n"], d["l2m"], d["l2numgroupsper"], d["l2acrossperc"]; rewirepercent=d["rewireperc"], numprocs=numprocs, seed=seed)

end

function gen_synthetic_non(yname::String; numprocs=1, seed=0)
	d = YAML.load_file(yname)
	return gen_synthetic(d; numprocs=numprocs, seed=seed)
end


function gen_er_non_from_non(non; seed=0)
	Random.seed!(seed)
	l1 = non
	l2 = non.upper

	newl1group = AbstractGraph[]
	for (Gind, G) in enumerate(l1.self.groups)
		n = nv(G)
		m = ne(G)
		seedval = rand(1:1000000000)
		newG = gen_er(n, m; seed=seedval)
		push!(newl1group, newG)
	end
	newl1GG = GraphOfGroups(newl1group, l1.self.tags)

	newl2group = AbstractGraph[]
	for (Gind, G) in enumerate(l2.self.groups)
		n = nv(G)
		m = ne(G)
		seedval = rand(1:1000000000)
		newG = gen_er(n, m; seed=seedval)
		push!(newl2group, newG)
	end
	newl2GG = GraphOfGroups(newl2group, l2.self.tags, l2.self.interedges)

	newl1 = NoN(newl1GG)
	newl2 = NoN(newl2GG)
	assign_lower!(newl2, newl1, l2.self2lower)
	return newl1
end
