using DelimitedFiles: writedlm
using LightGraphs
using LightGraphs: AbstractGraphFormat

import LightGraphs: loadgraph, loadgraphs, savegraph

using GraphIO

struct GMLFormat2 <: AbstractGraphFormat
    offset::Int
end
GMLFormat2() = GMLFormat2(0)

function _gml_read_one_graph(gs, dir)
    nodes = [x[:id] for x in gs[:node]]
    if dir
        g = LightGraphs.DiGraph(length(nodes))
    else
        g = LightGraphs.Graph(length(nodes))
    end
    mapping = Dict{Int,Int}()
    for (i, n) in enumerate(nodes)
        mapping[n] = i
    end
    sds = [(Int(x[:source]), Int(x[:target])) for x in gs[:edge]]
    for (s, d) in (sds)
        add_edge!(g, mapping[s], mapping[d])
    end
    return g
end

function loadgml(io::IO, gname::String)
    p = Parsers.GML.parse_dict(read(io, String))
    for gs in p[:graph]
        dir = Bool(get(gs, :directed, 0))
        graphname = get(gs, :label, dir ? "digraph" : "graph")

        (gname == graphname) && return _gml_read_one_graph(gs, dir)
    end
    error("Graph $gname not found")
end

function loadgml_mult(io::IO)
    p = Parsers.GML.parse_dict(read(io, String))
    graphs = Dict{String,LightGraphs.AbstractGraph}()
    for gs in p[:graph]
        dir = Bool(get(gs, :directed, 0))
        graphname = get(gs, :label, dir ? "digraph" : "graph")
        graphs[graphname] = _gml_read_one_graph(gs, dir)
    end
    return graphs
end

"""
    savegml(f, g, gname="graph")
Write a graph `g` with name `gname` to an IO stream `io` in the
[GML](https://en.wikipedia.org/wiki/Graph_Modelling_Language) format. Return 1.
"""
function savegml(io::IO, g::LightGraphs.AbstractGraph, gname::String = ""; offset=0)
    println(io, "graph")
    println(io, "[")
    length(gname) > 0 && println(io, "label \"$gname\"")
    is_directed(g) && println(io, "directed 1")
    for i = 1:nv(g)
        println(io, "\tnode")
        println(io, "\t[")
        println(io, "\t\tid $(i-offset)")
        println(io, "\t\tlabel \"$(i-offset)\"")
        println(io, "\t]")
    end
    for e in LightGraphs.edges(g)
        s, t = Tuple(e)
        println(io, "\tedge")
        println(io, "\t[")
        println(io, "\t\tsource $(s-offset)")
        println(io, "\t\ttarget $(t-offset)")
        println(io, "\t\tweight 1.0")
        println(io, "\t]")
    end
    println(io, "]")
    return 1
end


"""
    savegml_mult(io, graphs)
Write a dictionary of (name=>graph) to an IO stream `io` Return number of graphs written.
"""
function savegml_mult(io::IO, graphs::Dict; offset=0)
    ng = 0
    for (gname, g) in graphs
        ng += savegml(io, g, gname; offset=offset)
    end
    return ng
end
loadgraph(io::IO, gname::String, ::GMLFormat2) = loadgml(io, gname)
loadgraphs(io::IO, ::GMLFormat2) = loadgml_mult(io)
savegraph(io::IO, g::AbstractGraph, gname::String, format::GMLFormat2) = savegml(io, g, gname; offset=format.offset)
savegraph(io::IO, d::Dict, format::GMLFormat2) = savegml_mult(io, d; offset=format.offset)




struct EdgeListDelimFormat <: AbstractGraphFormat
    delim::String
end
EdgeListDelimFormat() = EdgeListDelimFormat("\t")

struct FlexibleEdgeListFormat <: AbstractGraphFormat
    delim::String
    hasheader::Bool
end
FlexibleEdgeListFormat() = FlexibleEdgeListFormat("\t", false)

function saveflexibleedgelist(io::IO, g::LightGraphs.AbstractGraph, gname::String; delim="\t", hasheader=false)
    # open(io, "w") do fo
        header = ""
        if hasheader
            header = "$(nv(g))$(delim)$(ne(g))\n"
            write(io, header)
        end
        for e in edges(g)
            offset = 0
            if hasheader
                offset = 1
            end
            write(io, string(src(e)-offset, delim, dst(e)-offset, "\n"))
        end

    # end
end
savegraph(io::IO, g::AbstractGraph, gname::String, felf::FlexibleEdgeListFormat) = saveflexibleedgelist(io, g, gname; delim=felf.delim, hasheader=felf.hasheader)

function saveedgelistdelim(io::IO, g::LightGraphs.AbstractGraph, gname::String; delim="\t")
    writedlm(io, ([src(e), dst(e)] for e in LightGraphs.edges(g)), delim)
    return 1
end

savegraph(io::IO, g::AbstractGraph, gname::String, eldf::EdgeListDelimFormat) = saveedgelistdelim(io, g, gname; delim=eldf.delim)

function write_labels_to_onehot(labelf)
    outf = labelf[1:end-4] * ".onehot.txt"
    distinct = []
    open(labelf) do fi
        for line in eachline(fi)
            if line in distinct
                continue
            end
            push!(distinct, line)
        end
    end
    numlabels = length(distinct)
    labels = distinct

    open(outf, "w") do fo
        open(labelf) do fi
            for line in eachline(fi)
                i = findfirst(isequal(line), labels)
                onehot = zeros(numlabels)
                onehot[i] = 1
                write(fo, join(onehot, " "), "\n")
            end
        end
    end

end

function savegraphofgroups(savedir, GG::GraphOfGroups, level, fillto; format=FlexibleEdgeListFormat(), prefix="")
    ext = ""

    if typeof(format) == FlexibleEdgeListFormat
        # println(typeof(format))
        if format.hasheader
            ext = ".sz_hdr.txt"
        else
            ext = ".txt"
        end
    else
        ext = ".gml.txt"
    end

    for (Gind, G) in enumerate(GG.groups)
        fname = string(prefix, "l", level, "_", lpad(Gind, fillto, "0"), ext)
        savegraph(string(savedir, fname), G, format)
    end
end

# function savenonlevel(savedir, non::AbstractNoN, level, fillto; delim="\t", prefix="")
#     currenttags = []
#     if !isdefined(non, :upper)
#         savegraphofgroups(savedir, non.self, level, fillto; delim="\t", prefix="sanity")
#         currenttags = non.self.tags
#         if length(non.self.interedges) != 0
#             G, nodetags = to_simple_graph(non.self)
#             savegraph(string(savedir, "l$(level).txt"), G, EdgeListDelimFormat(delim))
#         else
#             savegraphofgroups(savedir, non.self, level, fillto; delim="\t", prefix="")
#         end
#     else
#         uppernon = non.upper
#         upperself2lower = uppernon.self2lower
#         for gind in 1:length(uppernon.self.groups)
#             for node in vertices(uppernon.self.groups[gind])
#                 tup = (gind, node)
#                 if !(tup in keys(upperself2lower))
#                     continue
#                 end
#                 selfind = upperself2lower[tup]
#                 # println(selfind)
#                 fname = string(prefix, "l", level, "_", lpad(selfind, fillto, "0"), ".txt")
#                 savegraph(string(savedir, fname), non.self.groups[selfind], EdgeListDelimFormat(delim))
#             end
#         end
#         # savegraphofgroups(savedir, non.self, currentlevel, fillto; delim="\t")
#         currenttags = non.self.tags
#     end
#
#     return currenttags
# end

function save2levelnon(outdir, non::AbstractNoN; delim="\t", format=FlexibleEdgeListFormat())

    if typeof(format) == FlexibleEdgeListFormat
        # println(typeof(format))
        if format.hasheader
            ext = ".sz_hdr.txt"
        else
            ext = ".txt"
        end
        # println("is flex format, ext is $ext")
    else
        ext = ".gml.txt"
        # println("is gml format, ext is $ext")
    end

    if Sys.iswindows()
        l1dir = string(outdir, "\\l1\\")
        l2dir = string(outdir, "\\l2\\")
        if !ispath(outdir)
            run(`powershell mkdir $outdir`)
            
            run(`powershell mkdir $l1dir`)
            run(`powershell mkdir $l2dir`)
        end
        # run(`cmd /c if not exist $outdir mkdir $outdir`)
        # l1dir = string(outdir, "\\l1\\")
        # l2dir = string(outdir, "\\l2\\")
        # run(`cmd /c if not exist $l1dir mkdir $l1dir`)
        # run(`cmd /c if not exist $l2dir mkdir $l2dir`)
    else
        run(`mkdir -p $outdir`)
        l1dir = string(outdir, "/l1/")
        l2dir = string(outdir, "/l2/")
        run(`mkdir -p $l1dir`)
        run(`mkdir -p $l2dir`)
    end
    l2 = non.upper
    savegraphofgroups(l2dir, l2.self, 2, 2; prefix="sanity", format=format)

    l2G, nodetags = to_simple_graph(l2.self)
    # println(nv(l2G))
    # println(ne(l2G))
    println("saving $(string(l2dir, "l2", ext)) in format $(typeof(format))")
    savegraph(string(l2dir, "l2", ext), l2G, format)

    savegraphofgroups(l1dir, non.self, 1, length(string(nv(l2G))); format=format)
    open("$outdir/labels.txt", "w") do fo
        for (Gind, G) in enumerate(l2.self.groups)
            for (nind, node) in enumerate(vertices(G))
                l2tag = l2.self.tags[Gind]
                tup = (Gind, nind)
                l1tag = non.self.tags[l2.self2lower[tup]]
                write(fo, l1tag, "-", l2tag, "\n")
            end
        end
    end
    write_labels_to_onehot("$outdir/labels.txt")

end


function savenon(nondir::String, non::AbstractNoN; delim="\t")
    println("saving NoN")
    run(`mkdir -p $nondir`)
    currentlevel = 0
    currenttags = []
    prevtags = []
    while true
        totalnodes = nvg(non.self)
        fillto = length(string(totalnodes))
        currentlevel += 1
        currentdir = string(nondir, "/l$(currentlevel)/")
        run(`mkdir -p $(currentdir)`)

        currenttags = savenonlevel(currentdir, non, currentlevel, fillto; delim="\t", prefix="")

        # if length(non.self.interedges) != 0
        #     savegraphofgroups(currentdir, non.self, currentlevel, fillto; delim="\t", prefix="sanity_")
        #     G, nodetags = to_simple_graph(non.self)
        #     currenttags = non.self.tags
        #     savegraph(string(currentdir, "l$(currentlevel).txt"), G, EdgeListDelimFormat(delim))
        # else
        #     savegraphofgroups(currentdir, non.self, currentlevel, fillto; delim="\t")
        #     currenttags = non.self.tags
        # end

        if length(currenttags) > 0 && length(prevtags) > 0
            current_self2lower = non.self2lower
            open(string(currentdir, "labs.txt"), "w") do fo
                for gind in 1:length(non.self.groups)
                    for node in vertices(non.self.groups[gind])
                        tup = (gind, node)
                        if !(tup in keys(current_self2lower))
                            write(fo, "\n")
                            continue
                        end
                        prevtagind = current_self2lower[tup]
                        if !(prevtagind in 1:length(prevtags))
                            continue
                        end
                        write(fo, prevtags[prevtagind], "\t", non.self.tags[gind], "\n")
                    end
                end
            end
        end
        prevtags = currenttags
        if !isdefined(non, :upper)
            break
        end
        non = non.upper
    end

end
