using Test
using LightGraphs, SyntheticNoNs

@testset "InterEdge equality" begin
	inter0 = InterEdge((1, 400), (4, 382))
	inter1 = InterEdge((2, 100), (3, 125))
	@test inter0 != inter1
	inter2 = InterEdge((2, 100), (3, 125))
	@test inter1 == inter2
	inter3 = InterEdge((3, 125), (2, 100))
	@test inter2 == inter3
	inter4 = InterEdge((2, 100), (3, 125); is_directed=true)
	@test inter2 != inter4
	inter5 = InterEdge((2, 100), (3, 125); is_directed=true)
	@test inter4 == inter5
	inter6 = InterEdge((3, 125), (2, 100); is_directed=true)
	@test inter5 != inter6
end


@testset "Graph of groups" begin
	G1 = gen_geo(200, 800; seed=1)
	G2 = gen_geo(200, 800; seed=2)
	G3 = gen_sf(200, 800; seed=3)
	G4 = gen_sf(200, 800; seed=4)
	G1G2 = InterEdge((1, 1), (2, 1))
	Gs = [G1, G2, G3, G4]
	tags = ["geo", "geo", "sf", "sf"]
	GG = GraphOfGroups(Gs, tags, [G1G2])
	@test typeof(GG) == GraphOfGroups{SimpleGraph{Int64}}
	@test eltype(GraphOfGroups{SimpleGraph{Int64}}) == SimpleGraph{Int64}
	@test eltype(GG) == SimpleGraph{Int64}
	GGDi = GraphOfGroups([SimpleDiGraph(1)], ["di"])
	@test typeof(GGDi) == GraphOfGroups{SimpleDiGraph{Int64}}
	@test eltype(GraphOfGroups{SimpleDiGraph{Int64}}) == SimpleDiGraph{Int64}
	@test eltype(GGDi) == SimpleDiGraph{Int64}
	@test GG.interedges == [G1G2]
	@test_throws DimensionMismatch("Each group must have exactly one tag.") GraphOfGroups(Gs, [tags..., "sf"])
	dummyinter = InterEdge((5, 1), (2, 1))
	dummyinter2 = InterEdge((1, 500), (3, 4))
	dummyinter3 = InterEdge((1, 125), (1, 140))

	@test !add_interedge!(GG, dummyinter)
	@test !add_interedge!(GG, dummyinter2)
	@test !add_interedge!(GG, dummyinter3)
	inter1 = InterEdge((2, 100), (3, 125))
	@test add_interedge!(GG, inter1)
	@test GG.interedges == [G1G2, inter1]

	@test rem_interedge!(GG, inter1)
	@test !rem_interedge!(GG, dummyinter)
	@test rem_interedge!(GG, InterEdge((2, 1), (1, 1))) # reverse of G1G2, should be the same since not directed
	@test GG.interedges == []
end

@testset "Graph of groups to SimpleGraph" begin

	G1 = SimpleGraph(4)
	add_edge!(G1, 1, 2)
	add_edge!(G1, 2, 3)
	add_edge!(G1, 3, 4)
	add_edge!(G1, 4, 1)
	G2 = path_graph(4)
	G3 = path_graph(5)
	add_edge!(G3, 2, 5)
	inter1 = InterEdge((1, 1), (3, 5))
	inter2 = InterEdge((2, 4), (3, 1))
	inter3 = InterEdge((1, 2), (2, 3))
	GG = GraphOfGroups([G1, G2, G3], ["square", "4path", "5path*"], [inter1, inter2, inter3])
	expectedG = SimpleGraph(13)
	# G1
	@test add_edge!(expectedG, 1, 2)
	@test add_edge!(expectedG, 2, 3)
	@test add_edge!(expectedG, 3, 4)
	@test add_edge!(expectedG, 4, 1)
	# G2
	@test add_edge!(expectedG, 5, 6)
	@test add_edge!(expectedG, 6, 7)
	@test add_edge!(expectedG, 7, 8)
	# G3
	@test add_edge!(expectedG, 9, 10)
	@test add_edge!(expectedG, 10, 11)
	@test add_edge!(expectedG, 11, 12)
	@test add_edge!(expectedG, 12, 13)
	@test add_edge!(expectedG, 10, 13)
	# inter
	@test add_edge!(expectedG, 1, 13)
	@test add_edge!(expectedG, 8, 9)
	@test add_edge!(expectedG, 2, 7)

	combG, nodetags = to_simple_graph(GG)
	@test nv(combG) == length(nodetags)
	@test combG == expectedG
	@test nodetags == ["square", "square", "square", "square", "4path", "4path", "4path", "4path", "5path*", "5path*", "5path*", "5path*", "5path*"]

	G4 = path_graph(3)
	add_edge!(G4, 1, 3)
	add_graph!(GG, G4, "triangle")
	inter4 = InterEdge((3, 1), (4, 2))
	@test add_interedge!(GG, inter4)
	# G4
	@test add_vertices!(expectedG, 3) == 3
	@test add_edge!(expectedG, 14, 15)
	@test add_edge!(expectedG, 15, 16)
	@test add_edge!(expectedG, 16, 14)
	# inter4
	@test add_edge!(expectedG, 9, 15)

	combG, nodetags = to_simple_graph(GG)
	@test nv(combG) == length(nodetags)
	@test combG == expectedG
	@test nodetags == ["square", "square", "square", "square", "4path", "4path", "4path", "4path", "5path*", "5path*", "5path*", "5path*", "5path*", "triangle", "triangle", "triangle"]

	@test !rem_graph!(GG, 5)
	@test rem_graph!(GG, 2)
	combG, nodetags = to_simple_graph(GG)
	@test nv(combG) == length(nodetags)
	expectedG2 = SimpleGraph(12)
	# G1
	@test add_edge!(expectedG2, 1, 2)
	@test add_edge!(expectedG2, 2, 3)
	@test add_edge!(expectedG2, 4, 3)
	@test add_edge!(expectedG2, 1, 4)
	# G3
	@test add_edge!(expectedG2, 5, 6)
	@test add_edge!(expectedG2, 6, 7)
	@test add_edge!(expectedG2, 7, 8)
	@test add_edge!(expectedG2, 8, 9)
	@test add_edge!(expectedG2, 6, 9)
	# G4
	@test add_edge!(expectedG2, 10, 11)
	@test add_edge!(expectedG2, 11, 12)
	@test add_edge!(expectedG2, 12, 10)
	# inter
	@test add_edge!(expectedG2, 1, 9)
	@test add_edge!(expectedG2, 5, 11)

	@test combG == expectedG2
	@test nodetags == ["square", "square", "square", "square", "5path*", "5path*", "5path*", "5path*", "5path*", "triangle", "triangle", "triangle"]

end

@testset "NoN" begin

	# println(SimpleGraph{Int64} <: NoNTypes)
	# println(GraphOfGroups{SimpleGraph{Int64}} <: GraphOfGroups{AbstractGraph})
	l1G1 = barabasi_albert(100, 9; seed=0)
	l1G2 = barabasi_albert(100, 9; seed=1)

	l2G = path_graph(3)
	l2G2 = path_graph(4)
	add_edge!(l2G2, 4, 1)
	l2GG = GraphOfGroups([l2G, l2G2], ["3path", "square"], [InterEdge((0, 0), (0, 0))])
	l2 = NoN(l2GG)
	@test typeof(l2) == NoN{SimpleGraph{Int64}, GraphOfGroups{SimpleGraph{Int64}}}
	@test eltype(l2) == typeof(l2GG)
	l1G1 = path_graph(1)
	l1G2 = path_graph(2)
	l1G3 = path_graph(3)
	l1G4 = path_graph(4)
	l1G5 = path_graph(5)
	l1G6 = path_graph(6)
	l1G7 = path_graph(7)
	l1GG = GraphOfGroups([l1G1, l1G2, l1G3, l1G4, l1G5, l1G6, l1G7], ["node", "2path", "3path", "4path", "5path", "6path", "7path"])
	l1 = NoN(l1GG)
	assign_lower!(l2, l1, Dict((1, 1) => 1, (1, 2) => 2, (1, 3) => 3, (2, 1) => 4, (2, 2) => 5, (2, 3) => 6, (2, 4) => 7))
	# assign_upper!(l1, l2, Dict(1 => (1, 1), 2 => (1, 2), 3 => (1, 3)))
	@test !isdefined(l2, :upper)
	@test !isdefined(l1, :lower)
	@test l2.self2lower[(1, 1)] == 1
	@test l2.self2lower[(1, 2)] == 2
	@test l2.self2lower[(1, 3)] == 3
	@test l2.self2lower[(2, 1)] == 4
	@test l2.self2lower[(2, 2)] == 5
	@test l2.self2lower[(2, 3)] == 6
	@test l2.self2lower[(2, 4)] == 7
	@test l1.self2upper[1] == (1, 1)
	@test l1.self2upper[2] == (1, 2)
	@test l1.self2upper[3] == (1, 3)
	@test l1.self2upper[4] == (2, 1)
	@test l1.self2upper[5] == (2, 2)
	@test l1.self2upper[6] == (2, 3)
	@test l1.self2upper[7] == (2, 4)
	# @test l2.nl == 2
	# @test l1.nl == 2
	# add_edge!(l1.self.groups[3], 1, 3)
	# println(collect(edges(l2.lower.self.groups[3])))
	# savenon("../data/toy-matching-nums/", l1; delim="\t")


	l2G = path_graph(3)
	l2G2 = path_graph(4)
	add_edge!(l2G2, 4, 1)
	l2GG = GraphOfGroups([l2G, l2G2], ["3path", "square"], [InterEdge((0, 0), (0, 0))])
	l2 = NoN(l2GG)

	l1G1 = path_graph(1)
	l1G2 = path_graph(2)
	l1G3 = path_graph(3)
	l1G4 = path_graph(4)
	l1G5 = path_graph(5)
	l1G6 = path_graph(6)
	l1G7 = path_graph(7)
	l1GG = GraphOfGroups([l1G1, l1G2, l1G3, l1G5, l1G6, l1G7], ["node", "2path", "3path", "5path", "6path", "7path"])
	l1 = NoN(l1GG)
	assign_lower!(l2, l1, Dict((1, 1) => 1, (1, 2) => 2, (1, 3) => 3, (2, 2) => 4, (2, 3) => 5, (2, 4) => 6))
	# savenon("../data/toy-more-l2-than-l1/", l1; delim="\t")

	l2G = path_graph(3)
	l2G2 = path_graph(4)
	add_edge!(l2G2, 4, 1)
	l2GG = GraphOfGroups([l2G, l2G2], ["3path", "square"], [InterEdge((0, 0), (0, 0))])
	l2 = NoN(l2GG)

	l1G1 = path_graph(1)
	l1G2 = path_graph(2)
	l1G3 = path_graph(3)
	l1G4 = path_graph(4)
	l1G5 = path_graph(5)
	l1G6 = path_graph(6)
	l1G7 = path_graph(7)
	l1G8 = path_graph(8)
	l1GG = GraphOfGroups([l1G1, l1G2, l1G3, l1G4, l1G5, l1G6, l1G7, l1G8], ["node", "2path", "3path", "4path", "5path", "6path", "7path", "8path"])
	l1 = NoN(l1GG)
	assign_lower!(l2, l1, Dict((1, 1) => 1, (1, 2) => 2, (1, 3) => 3, (2, 1) => 4, (2, 2) => 5, (2, 3) => 6, (2, 4) => 8))
	# savenon("../data/toy-more-l1-than-l2/", l1; delim="\t")
end

@testset "NoN gen" begin

end
