using ArgParse
using LightGraphs
using Flux.Data: DataLoader
using Flux: onecold, logitcrossentropy, @epochs, train!, Dense, Chain, ADAM, params, update!, gpu, cpu, softmax
import Zygote: Params, gradient
# println("not loading Flux until necessary")
push!(LOAD_PATH, "../")
using LinearAlgebra
using Statistics
using StatsBase
using Parameters: @with_kw
using SyntheticNoNs
using YAML
using CUDA
if has_cuda()
    @info "CUDA is on"
    CUDA.allowscalar(false)
end
push!(LOAD_PATH, "./scripts/")
using ProteinNoNLoader
include("utils.jl")
println("finished loading packages")

# julia ppiscripts/log_reg.jl --negclass negbp --network _induced --numnodes 4685

function parsecli()
	apsettings = ArgParseSettings()
	@add_arg_table apsettings begin
		"--basedir", "-b"
			default = "/scratch/sgu3/cleannondata/"
			help = "location of the data. The 'l1' and 'l2' folders should appear in this directory"
		"--outdir", "-o"
			default = "/scratch/sgu3/cleannondata/"
			help = "directory of output."
		"--goparams", "-g"
			default = "Homo_sapiens-protein-GO.EXP.bp.txt"
			help = "constraints on the GO terms to use. Here, biological process experimentally inferred annotations."
		"--classcutoff", "-c"
			arg_type = Int
			default = 20
			help = "minumum size of positive class"
		"--lr", "-l"
			arg_type = Float64
			default = 0.01
			help = "learning rate"
		"--epochs", "-e"
			arg_type = Int
			default = 100
		"--negclass", "-n"
			default = "negbp"
			help = "type of negative class. Here, proteins that are not annotated by any biological process GO terms"
		"--numnodes"
			arg_type = Int
			default = 18708
			help = "number of nodes in the ppi network"
		"--mode"
			default = ""
			help = "type of feature to use"
	end
	return parse_args(ARGS, apsettings)
end
@with_kw mutable struct Args
    η::Float64 = 0.01
    batchsize::Int = 1024
    epochs::Int = 10
    device::Function = gpu
end
Base.show(io::IO, args::Args) = print(io, "Learning rate: $(args.η)\nBatch size: $(args.batchsize)\nEpochs: $(args.epochs)\nDevice: $(args.device)")


parsedargs = parsecli()
println(parsedargs)
u_basedir = parsedargs["basedir"]
u_outdir = parsedargs["outdir"]
u_goparams = parsedargs["goparams"]
u_classcutoff = parsedargs["classcutoff"]
u_lr = parsedargs["lr"]
u_epochs = parsedargs["epochs"]
u_negclass = parsedargs["negclass"]
u_networktype = ""
u_godir = "$u_outdir/$(u_goparams[1:end-4])-$(u_classcutoff)-$(u_negclass)$u_networktype"
u_resoutdir = "$u_outdir/$(u_goparams[1:end-4])-$(u_classcutoff)-$(u_negclass)$(u_networktype)-results/"
run(`mkdir -p $u_resoutdir`)
u_l2n = parsedargs["numnodes"]
u_mode = parsedargs["mode"]


# # calculate accuracy from input features
function accuracy(x, y, model, mask)
    ypred = model(x)
	return accuracy(ypred, y, mask)
end

# # calculate accuracy from predictions
function accuracy(ypred, y, mask)
	acc = 0
	for i in 1:size(y)[2]
		# # only consider entities that are for the given training/val/testing set
        if mask[i]
            acc += sum(onecold(ypred[:, i]) == onecold(y[:, i]))
        end
    end
    return acc/sum(mask)
end

# # training function for Flux. See https://fluxml.ai/Flux.jl/stable/training/training/ for more information.
function customtrain!(losstrain, lossval, ps, feats, labels, opt; cb= () -> ())
    t0 = time()
    ps = Params(ps)

    try
        gs = gradient(ps) do
            losstrain(feats, labels)
        end
        update!(opt, ps, gs)
        cb()
    catch ex
        rethrow(ex)
    end
    println("Training loss: $(round(losstrain(feats, labels); digits=4))")
    println("Val loss: $(round(lossval(feats, labels); digits=4))")
    println("Epoch time: $(time() - t0)")
end

# # args: hyperparams
# # outdir: unused
# # resoutdir: directory where results should be written
# # labeldir: directory containing GO terms and which proteins are positive/negative data instances
# # mode: approach
# # l2numnodes: number of nodes in the level 2 network
# # numprocs: number of processes to use
# # seedrange: which random seeds to use for the training/val/testing split
function log_reg_pipeline(args, outdir, resoutdir, labeldir, mode, l2numnodes; numprocs=1, seedrange=1:3)

	# # data structure of the features
	feats = Array{Array{Float64, 1}, 1}()
	feats = loadfeats(outdir, l2numnodes, mode; numprocs=numprocs)
	origfeats = deepcopy(feats)
	
	# # loop through all 131 GO terms considered
	for golabelf in readdir(labeldir)
		goterm = golabelf[1:end-4]
		if occursin("sizes", golabelf)
			continue
		end
		println("binary classification for $golabelf")



		# # load labels, which are one-hot encodings.
		labels = loadlabels("$labeldir/$golabelf")
		origlabels = deepcopy(labels)

		# # random sample the training/val/testing split in a 1:1:1 ratio
		trainmasks = Vector{Array{Bool, 2}}()
	    valmasks = Vector{Array{Bool, 2}}()
	    testmasks = Vector{Array{Bool, 2}}()
	    for seed in seedrange
	        trainmask, valmask, testmask = gen_train_val_test_masks(feats, labels, 1/3; seed=seed)
	        push!(trainmasks, trainmask)
	        push!(valmasks, valmask)
	        push!(testmasks, testmask)
	    end
	    trainaccs = []
	    valaccs = []
	    testaccs = []
		for seed in seedrange


			# # create directory for 'log' file (just keeps track of if classification has already been ran for this GO term)
			logdir = "$resoutdir/logs/"
			run(`mkdir -p $logdir`)
			logf = "$logdir/$(golabelf)_lr$(args.η)_$(mode)_seed$(seed).log"

			# # create directory for results for each seed (prediction probabilities, masks, etc.)
			indoutdir = "$resoutdir/indres/"
			run(`mkdir -p $indoutdir`)

			# # name of results file
			resultsfname = "$indoutdir/$(golabelf)_lr$(args.η)_$(mode)_seed$(seed).results"
			predf = "$resultsfname.preds"
			valpredf = "$resultsfname.val.preds"

			# # check if log exists, if it does, read information from it and continue
			if isfile(logf)
				println("already made $logf, skipping")
				open(resultsfname, "r") do fi
					for line in eachline(fi)
						res = 0.0
						if occursin("Training", line)
							res = parse(Float64, split(line)[3])
							push!(trainaccs, res)
						elseif occursin("Val", line)
							res = parse(Float64, split(line)[3])
							push!(valaccs, res)
						elseif occursin("Test", line)
							res = parse(Float64, split(line)[3])
							push!(testaccs, res)
						end
					end
				end
				continue
			end

			# # split data
			trainmask = trainmasks[seed]
	        valmask = valmasks[seed]
	        testmask = testmasks[seed]

	        # # save mask data for reproducibility
			maskdir = "$resoutdir/masks/"
			run(`mkdir -p $maskdir`)
	        writemask(trainmask, "$maskdir/$(golabelf[1:end-4])_trainmask$seed.txt")
	        writemask(valmask, "$maskdir/$(golabelf[1:end-4])_valmask$seed.txt")
	        writemask(testmask, "$maskdir/$(golabelf[1:end-4])_testmask$seed.txt")

			feats = origfeats
			labels = origlabels

			# # format feats
	        feats = reduce(hcat, feats)
	        labels = reduce(hcat, labels)

	        # # record sizes for model
			featsize = size(feats)[1]
	        numlabels = size(labels)[1]

	        # # create log reg model
	        model = Chain(Dense(featsize, numlabels), softmax)

			optimiser = ADAM(args.η)

			# # move to GPU if available
			model = args.device(model)
	        feats = args.device(feats)
	        labels = args.device(labels)

			trainmask = args.device(trainmask)
	        valmask = args.device(valmask)

	        # # loss functions using masks instead of actual disjoin training/val/test sets.
	        # # loss functions only see information from training/validation data
			function losstrain(x, y)
	            ypred = model(x)
	            return logitcrossentropy(ypred .* trainmask, y .* trainmask)
	        end
	        function lossval(x, y)
	            ypred = model(x)
	            return logitcrossentropy(ypred .* valmask, y .* valmask)
	        end

	        # # some extra info during training iterations
			function callback()
				println("Training acc: ", accuracy(feats, labels, model, trainmask))
				println("Val acc: ",  accuracy(feats, labels, model, valmask))
			end
			println("Training start.")
	        @epochs args.epochs customtrain!(losstrain, lossval, params(model), feats, labels, optimiser; cb=() -> callback())

	        # # move data back to cpu for calculating accuracy
			model = cpu(model)
	        feats = cpu(feats)
	        labels = cpu(labels)
			trainmask = cpu(trainmask)
	        valmask = cpu(valmask)

			trainacc = accuracy(feats, labels, model, trainmask)
	        push!(trainaccs, trainacc)
	        valacc = accuracy(feats, labels, model, valmask)
	        push!(valaccs, valacc)
	        testacc = accuracy(feats, labels, model, testmask)
	        push!(testaccs, testacc)


	        # # write prediction probabilities to file
			valpreds = model(feats) .* valmask
			open(valpredf, "w") do fo
				for i in 1:size(valpreds)[2]
					println(fo, valpreds[:, i])
				end
			end

			preds = model(feats) .* testmask
			open(predf, "w") do fo
				for i in 1:size(preds)[2]
					println(fo, preds[:, i])
				end
			end

			# # record accuracies (not actually used for biological NoNs)
	        open(resultsfname, "w") do fo
	            println(fo, args)
	            write(fo, "Training accuracy: $(trainacc)\n")
	            write(fo, "Val accuracy: $(valacc)\n")
	            write(fo, "Test accuracy: $(testacc)\n")
	        end

			open(logf, "w") do logfo
				println(logfo, "ran this already")
			end


		end
		resultsavgf = "$resoutdir/$(golabelf[1:end-4])_lr$(args.η)_$(mode).results.avg"
		open(resultsavgf, "w") do fo
			println(fo, "Training accuracy: $(mean(trainaccs)) ± $(std(trainaccs)), $(trainaccs)")
			println(fo, "Val accuracy: $(mean(valaccs)) ± $(std(valaccs)), $(valaccs)")
			println(fo, "Test accuracy: $(mean(testaccs)) ± $(std(testaccs)), $(testaccs)")
		end
		println("Training accuracy: $(mean(trainaccs)) ± $(std(trainaccs))")
		println("Val accuracy: $(mean(valaccs)) ± $(std(valaccs))")
		println("Test accuracy: $(mean(testaccs)) ± $(std(testaccs))")

		println("only doing the first GO term for now, comment out the following 'break' to run for all GO terms")
		break;
	end
end

# # hyperparams
u_args = Args(η=u_lr, epochs=u_epochs)

log_reg_pipeline(u_args, u_outdir, u_resoutdir, u_godir, u_mode, u_l2n; numprocs=1, seedrange=1:3)
