using ArgParse
using LightGraphs
using Flux.Data: DataLoader
using Flux: onecold, logitcrossentropy, @epochs, train!, Dense, Chain, ADAM, params, update!, gpu, cpu, softmax
import Zygote: Params, gradient
push!(LOAD_PATH, "../")
using FluxGNNs
using LinearAlgebra
using SparseArrays
using Statistics
using StatsBase
using Parameters: @with_kw
using SyntheticNoNs
using YAML
using CUDA
if has_cuda()
    @info "CUDA is on"
    CUDA.allowscalar(false)
end
push!(LOAD_PATH, "./scripts/")
using ProteinNoNLoader
include("utils.jl")
println("finished loading packages")

function parsecli()
	apsettings = ArgParseSettings()
	@add_arg_table apsettings begin
		"--basedir", "-b"
			default = "/scratch/sgu3/cleannondata/"
		"--outdir", "-o"
			default = "/scratch/sgu3/cleannondata/"
		"--goparams", "-g"
			default = "Homo_sapiens-protein-GO.EXP.bp.txt"
		"--classcutoff", "-c"
			arg_type = Int
			default = 20
		"--lr", "-l"
			arg_type = Float64
			default = 0.01
		"--epochs", "-e"
			arg_type = Int
			default = 100
		"--angstromcutoff", "-a"
			arg_type = Float64
			default = 6.0
		"--negclass", "-n"
			default = "negbp"
		"--numnodes"
			arg_type = Int
			default = 18708
		"--mode"
			default = ""
	end
	return parse_args(ARGS, apsettings)
end
@with_kw mutable struct Args
    η::Float64 = 0.01
    batchsize::Int = 1024
    epochs::Int = 10
    device::Function = gpu
	channels::Int = 256
end
Base.show(io::IO, args::Args) = print(io, "Learning rate: $(args.η)\nBatch size: $(args.batchsize)\nEpochs: $(args.epochs)\nDevice: $(args.device)")

parsedargs = parsecli()
println(parsedargs)
u_basedir = parsedargs["basedir"]
u_outdir = parsedargs["outdir"]
u_goparams = parsedargs["goparams"]
u_classcutoff = parsedargs["classcutoff"]
u_angstromcutoff = parsedargs["angstromcutoff"]
u_lr = parsedargs["lr"]
u_epochs = parsedargs["epochs"]
u_negclass = parsedargs["negclass"]
u_networktype = ""



u_godir = "$u_outdir/$(u_goparams[1:end-4])-$(u_classcutoff)-$(u_negclass)$u_networktype"
u_resoutdir = "$u_outdir/$(u_goparams[1:end-4])-$(u_classcutoff)-$(u_negclass)$(u_networktype)-results/"
run(`mkdir -p $u_resoutdir`)

u_l2n = parsedargs["numnodes"]
u_mode = parsedargs["mode"]

u_l1dir = "$u_outdir/l1/"
u_l2dir = "$u_outdir/l2/"
u_procdir = "$u_outdir/processed/"
run(`mkdir -p $u_procdir`)

include("bio_non_gcn_helper.jl")




function accuracy(x, y, model, mask)
    ypred = model(x)
	return accuracy(ypred, y, mask)
end

function accuracy(ypred, y, mask)
	acc = 0
	for i in 1:size(y)[1]
        if mask[i]
            acc += sum(onecold(ypred[i, :]) == onecold(y[i, :])) #
        end
    end
    return acc/sum(mask)
end

function customtrain!(losstrain, lossval, ps, feats, labels, opt; cb= () -> ())
    t0 = time()
    ps = Params(ps)

    try
        gs = gradient(ps) do
            losstrain(feats, labels)
        end
        update!(opt, ps, gs)
        cb()
    catch ex
        rethrow(ex)
    end
    println("Val loss: $(round(lossval(feats, labels); digits=4))")
    println("Epoch time: $(time() - t0)")
end

function procgcn_pipeline(args, l1dir, l2dir, procdir, resoutdir, oldgodir, labeldir, l2numnodes, mode; numprocs=1, seedrange=1:3)

	G, feats = makeproc2(l1dir, l2dir, procdir, l2numnodes)
	feats = reduce(hcat, feats)'
	adj = adjacency_matrix(G)
	println("adjmat size $(size(adj))")


	println("preprocessing adj")
	adjhat = preprocess(adj)
	println("finished preprocessing")
	procf = "$procdir/l1l2proc2.sz_hdr.txt"
	remakelabels(oldgodir, labeldir, procf)

	hashcattedfeats = false
	hashcattedlabs = false
	golabind = 0
	for golabelf in readdir(labeldir)
		golabind += 1
		
		if occursin("sizes", golabelf)
			continue
		end
		println("binary classification for $golabelf")

		labels = loadlabels("$labeldir/$golabelf")

		trainmasks = Vector{Array{Bool, 2}}()
	    valmasks = Vector{Array{Bool, 2}}()
	    testmasks = Vector{Array{Bool, 2}}()
	    for seed in seedrange
	        trainmask, valmask, testmask = gen_train_val_test_masks(feats, labels, 1/3; seed=seed)
	        push!(trainmasks, trainmask)
	        push!(valmasks, valmask)
	        push!(testmasks, testmask)
	    end
	    trainaccs = []
	    valaccs = []
	    testaccs = []
		for seed in seedrange

			logdir = "$resoutdir/logs/"
			if Sys.iswindows()
				if !ispath(logdir)
					run(`powershell mkdir $logdir`)
				end
			else
				run(`mkdir -p $logdir`)
			end
			logf = "$logdir/$(golabelf)_lr$(args.η)_$(mode)_seed$(seed).log"

			indoutdir = "$resoutdir/indres/"
			if Sys.iswindows()
				if !ispath(indoutdir)
					run(`powershell mkdir $indoutdir`)
				end
			else
				run(`mkdir -p $indoutdir`)
			end
			resultsfname = "$indoutdir/$(golabelf)_lr$(args.η)_$(mode)_seed$(seed).results"
			predf = "$resultsfname.preds"
			valpredf = "$resultsfname.val.preds"
			if isfile(logf)
				println("already made $logf, skipping")
				open(resultsfname, "r") do fi
					for line in eachline(fi)
						res = 0.0
						if occursin("Training", line)
							res = parse(Float64, split(line)[3])
							push!(trainaccs, res)
						elseif occursin("Val", line)
							res = parse(Float64, split(line)[3])
							push!(valaccs, res)
						elseif occursin("Test", line)
							res = parse(Float64, split(line)[3])
							push!(testaccs, res)
						end

					end
				end
				continue
			end


			trainmask = trainmasks[seed]'
	        valmask = valmasks[seed]'
	        testmask = testmasks[seed]'

			maskdir = "$procdir/masks/"
			if Sys.iswindows()
				if !(ispath(maskdir))
					run(`powershell mkdir $maskdir`)
				end
			else
				run(`mkdir -p $maskdir`)
			end

	        writemask2(trainmask, "$maskdir/$(golabelf[1:end-4])_trainmask$seed.txt"; numlines=l2numnodes)
	        writemask2(valmask, "$maskdir/$(golabelf[1:end-4])_valmask$seed.txt"; numlines=l2numnodes)
	        writemask2(testmask, "$maskdir/$(golabelf[1:end-4])_testmask$seed.txt"; numlines=l2numnodes)

			

	        if seed == seedrange[1]
				labels = reduce(hcat, labels)'
	        end

			featsize = size(feats)[2]
	        numlabels = size(labels)[2]
			println("featsize: $featsize")
			println("numlabels: $numlabels")


			gcl1 = GCLayer(featsize, args.channels, adjhat)
			gcl2 = GCLayer(args.channels, args.channels, adjhat)
			lrlayer = DenseEnd(args.channels, numlabels)
			customsoftmax(x) = softmax(x; dims=2)
			function customsoftmax(x)
				exp.(x) ./ sum(exp.(x), dims=2)
			end
	        model = Chain(gcl1, gcl2, lrlayer)
			println("chained the layers together")




			optimiser = ADAM(args.η)

			model = args.device(model)
	        feats = args.device(feats)
	        labels = args.device(labels)

			trainmask = args.device(trainmask)
	        valmask = args.device(valmask)


			function losstrain(x, y)
	            ypred = model(x)
	            return logitcrossentropy(ypred .* trainmask, y .* trainmask)
	        end
	        function lossval(x, y)
	            ypred = model(x)
	            return logitcrossentropy(ypred .* valmask, y .* valmask)
	        end



			function callback()

			end
			println("Training start.")
	        @epochs args.epochs customtrain!(losstrain, lossval, params(model), feats, labels, optimiser; cb=() -> callback())

			model = cpu(model)
	        feats = cpu(feats)
	        labels = cpu(labels)
			trainmask = cpu(trainmask)
	        valmask = cpu(valmask)

			trainacc = accuracy(feats, labels, model, trainmask)
	        push!(trainaccs, trainacc)
	        valacc = accuracy(feats, labels, model, valmask)
	        push!(valaccs, valacc)
	        testacc = accuracy(feats, labels, model, testmask)
	        push!(testaccs, testacc)


			valpreds = model(feats) .* valmask
			open(valpredf, "w") do fo
				for i in 1:l2numnodes
					println(fo, valpreds[i, :])
				end
			end

			preds = model(feats) .* testmask
			open(predf, "w") do fo
				for i in 1:l2numnodes
					println(fo, preds[i, :])
				end
			end

	        open(resultsfname, "w") do fo
	            println(fo, args)
	            write(fo, "Training accuracy: $(trainacc)\n")
	            write(fo, "Val accuracy: $(valacc)\n")
	            write(fo, "Test accuracy: $(testacc)\n")
	        end

			open(logf, "w") do logfo
				println(logfo, "ran this already")
			end
			println("currently not copying logfile over")

		end
		resultsavgf = "$resoutdir/$(golabelf[1:end-4])_lr$(args.η)_$(mode).results.avg"
		open(resultsavgf, "w") do fo
			println(fo, "Training accuracy: $(mean(trainaccs)) ± $(std(trainaccs)), $(trainaccs)")
			println(fo, "Val accuracy: $(mean(valaccs)) ± $(std(valaccs)), $(valaccs)")
			println(fo, "Test accuracy: $(mean(testaccs)) ± $(std(testaccs)), $(testaccs)")
		end
		println("Training accuracy: $(mean(trainaccs)) ± $(std(trainaccs))")
		println("Val accuracy: $(mean(valaccs)) ± $(std(valaccs))")
		println("Test accuracy: $(mean(testaccs)) ± $(std(testaccs))")

		println("only doing the first GO term for now, comment out the following 'break' to run for all GO terms")
		break;
	end
end


u_newgodir = "$u_procdir/$(u_goparams[1:end-4])-$(u_classcutoff)-$(u_negclass)$(u_networktype)2"

u_args = Args(η=0.001, epochs=u_epochs, channels=256, device=gpu)

# # only one seed for testing purposes since it takes longer
procgcn_pipeline(u_args, u_l1dir, u_l2dir, u_procdir, u_resoutdir, u_godir, u_newgodir, u_l2n, "convs2"; numprocs=1, seedrange=1:1)

