# # works similarly as bio_log_reg.jl

using ArgParse
using LightGraphs
using Flux.Data: DataLoader
using Flux: onecold, logitcrossentropy, @epochs, train!, Dense, Chain, ADAM, params, update!, gpu, cpu, softmax
import Zygote: Params, gradient
println("loaded flux")
push!(LOAD_PATH, "../")
using FluxGNNs
using LinearAlgebra
using SparseArrays
using Statistics
using StatsBase
using Parameters: @with_kw
using SyntheticNoNs
using YAML
using CUDA
if has_cuda()
    @info "CUDA is on"
    CUDA.allowscalar(false)
end
push!(LOAD_PATH, "./scripts/")
using ProteinNoNLoader
include("utils.jl")
println("finished loading packages")

function parsecli()
	apsettings = ArgParseSettings()
	@add_arg_table apsettings begin
		"--basedir", "-b"
			default = "/scratch/sgu3/cleannondata/"
		"--outdir", "-o"
			default = "/scratch/sgu3/cleannondata/"
		"--goparams", "-g"
			default = "Homo_sapiens-protein-GO.EXP.bp.txt"
		"--classcutoff", "-c"
			arg_type = Int
			default = 20
		"--lr", "-l"
			arg_type = Float64
			default = 0.01
		"--epochs", "-e"
			arg_type = Int
			default = 100
		"--angstromcutoff", "-a"
			arg_type = Float64
			default = 6.0
		"--negclass", "-n"
			default = "negbp"
		"--numnodes"
			arg_type = Int
			default = 18708
		"--mode"
			default = ""
	end
	return parse_args(ARGS, apsettings)
end
@with_kw mutable struct Args
    η::Float64 = 0.01
    batchsize::Int = 1024
    epochs::Int = 10
    device::Function = gpu
end
Base.show(io::IO, args::Args) = print(io, "Learning rate: $(args.η)\nBatch size: $(args.batchsize)\nEpochs: $(args.epochs)\nDevice: $(args.device)")


parsedargs = parsecli()
println(parsedargs)
u_basedir = parsedargs["basedir"]
u_outdir = parsedargs["outdir"]
u_goparams = parsedargs["goparams"]
u_classcutoff = parsedargs["classcutoff"]
u_angstromcutoff = parsedargs["angstromcutoff"]
u_lr = parsedargs["lr"]
u_epochs = parsedargs["epochs"]
u_negclass = parsedargs["negclass"]
u_networktype = ""
u_godir = "$u_outdir/$(u_goparams[1:end-4])-$(u_classcutoff)-$(u_negclass)$u_networktype"
u_resoutdir = "$u_outdir/$(u_goparams[1:end-4])-$(u_classcutoff)-$(u_negclass)$(u_networktype)-results/"
run(`mkdir -p $u_resoutdir`)
u_l2n = parsedargs["numnodes"]
u_mode = parsedargs["mode"]




function preprocess(adjmat::AbstractArray)
    adjmathat = adjmat
    degrees = sum(adjmathat, dims=2)
    degrees = vec(degrees) .^ -0.5
    normD = spdiagm(0 => degrees)
    return normD * adjmathat * normD
end

function get_tri_adjmat(adjlistf, adjmat)

	tempf = "$(adjlistf[1:end-4]).mtx"
	outf = "$tempf.micro"

	if !isfile(outf)
		edgecount = 0
		nodeset = Set()
		open(adjlistf, "r") do fi
			for line in eachline(fi)
				larr = split(line)
				node1 = larr[1]
				node2 = larr[2]
				push!(nodeset, node1)
				push!(nodeset, node2)
				edgecount += 1
			end
		end

		nnodes = length(nodeset)
		open(tempf, "w") do fo
			println(fo, nnodes, "\t", nnodes, "\t", edgecount)
			write(fo, read(adjlistf, String))
		end


		open("junk.txt", "w") do fo
			write(fo, read(`/afs/crc.nd.edu/user/s/sgu3/Private/non-embedding/utils/pgd -f $tempf --micro $outf`))
		end
	end

	triadjmat = deepcopy(adjmat)
	open(outf, "r") do fi
		readline(fi)
		for line in eachline(fi)
			larr = parse.(Int, split(line, ","))
			n1 = larr[1]
			n2 = larr[2]
			numtris = larr[3]
			triadjmat[n1, n2] = numtris * triadjmat[n1, n2]
			triadjmat[n2, n1] = numtris * triadjmat[n2, n1]
		end
	end
	run(`rm -f $tempf`)


	return triadjmat
end




function accuracy(x, y, model, mask)

    ypred = model(x)
	return accuracy(ypred, y, mask)

end

function accuracy(ypred, y, mask)
	acc = 0
	for i in 1:size(y)[1]
        if mask[i]
			acc += sum(onecold(ypred[i, :]) == onecold(y[i, :]))
        end
    end
    return acc/sum(mask)
end

function customtrain!(losstrain, lossval, ps, feats, labels, opt; cb= () -> ())
    t0 = time()
    ps = Params(ps)

    try
        gs = gradient(ps) do
            losstrain(feats, labels)
        end
        update!(opt, ps, gs)
        cb()
    catch ex
        rethrow(ex)
    end
    println("Training loss: $(round(losstrain(feats, labels); digits=4))")
    println("Val loss: $(round(lossval(feats, labels); digits=4))")
    println("Epoch time: $(time() - t0)")
end

function sign_pipeline(args, outdir, resoutdir, labeldir, mode, l2numnodes, ppow, spow, tpow; numprocs=1, seedrange=1:3)

	writeaverage = true



	G = loadedgelist("$outdir/l2/l2.txt")
	adj = adjacency_matrix(G)
	adjhat = preprocess(adj + sparse(I, size(adj)))

	featholder = []
	for i in 1:ppow
		push!(featholder, adjhat ^ i)
	end

	triadjmat = get_tri_adjmat("$outdir/l2/l2.txt", adj)
	triadjmat = preprocess(triadjmat + sparse(I, size(triadjmat)))
	for i in 1:tpow
		push!(featholder, triadjmat ^ i)
	end

	l1dir = "$outdir/l1/"
	pool = WorkerPool(collect(1:numprocs))
	feats = reduce(hcat, featholder)

	for golabelf in readdir(labeldir)

		if occursin("sizes", golabelf)
			continue
		end
		println("binary classification for $golabelf")
		
		logdir = "$resoutdir/logs/"
		run(`mkdir -p $logdir`)
		whichseeds = []
		for seed in seedrange
			logf = "$logdir/$(golabelf)_lr$(args.η)_$(mode)_seed$(seed).log"
			if isfile(logf)
				push!(whichseeds, seed)
			end
		end
		if whichseeds == collect(seedrange)
			println("already ran for this")
			continue
		end

		writeaverage = true
		goterm = golabelf[1:end-4]


		labels = loadlabels("$labeldir/$golabelf")
		origlabels = deepcopy(labels)

		trainmasks = Vector{Array{Bool, 2}}()
	    valmasks = Vector{Array{Bool, 2}}()
	    testmasks = Vector{Array{Bool, 2}}()
	    for seed in seedrange
	        trainmask, valmask, testmask = gen_train_val_test_masks(feats, labels, 1/3; seed=seed)
	        push!(trainmasks, trainmask)
	        push!(valmasks, valmask)
	        push!(testmasks, testmask)
	    end
	    trainaccs = []
	    valaccs = []
	    testaccs = []
		for seed in seedrange

			logdir = "$resoutdir/logs/"
			run(`mkdir -p $logdir`)
			logf = "$logdir/$(golabelf)_lr$(args.η)_$(mode)_seed$(seed).log"

			indoutdir = "$resoutdir/indres/"
			run(`mkdir -p $indoutdir`)
			resultsfname = "$indoutdir/$(golabelf)_lr$(args.η)_$(mode)_seed$(seed).results"
			predf = "$resultsfname.preds"
			valpredf = "$resultsfname.val.preds"

			if isfile(logf)
				writeaverage=false
				println("already made $logf, skipping")
				open(resultsfname, "r") do fi
					for line in eachline(fi)
						res = 0.0
						if occursin("Training", line)
							res = parse(Float64, split(line)[3])
							push!(trainaccs, res)
						elseif occursin("Val", line)
							res = parse(Float64, split(line)[3])
							push!(valaccs, res)
						elseif occursin("Test", line)
							res = parse(Float64, split(line)[3])
							push!(testaccs, res)
						end

					end
				end
				continue
			end


			trainmask = trainmasks[seed]'
	        valmask = valmasks[seed]'
	        testmask = testmasks[seed]'

			maskdir = "$resoutdir/masks/"
			run(`mkdir -p $maskdir`)
	        writemask(trainmask, "$maskdir/$(golabelf[1:end-4])_trainmask$seed.txt")
	        writemask(valmask, "$maskdir/$(golabelf[1:end-4])_valmask$seed.txt")
	        writemask(testmask, "$maskdir/$(golabelf[1:end-4])_testmask$seed.txt")


			labels = origlabels

	        labels = reduce(hcat, labels)'

			println("feats $(size(feats)), $(typeof(feats))")
			println("labels $(size(labels))")
			println("masks $(size(trainmask)), $(size(valmask))")


			featsize = size(feats)[2]
	        numlabels = size(labels)[2]
			lrlayer = DenseEnd(featsize, numlabels)
	        model = Chain(lrlayer, x->softmax(x, dims=2))

			optimiser = ADAM(args.η)

			model = args.device(model)
	        feats = args.device(feats)
	        labels = args.device(labels)

			trainmask = args.device(trainmask)
	        valmask = args.device(valmask)

			function losstrain(x, y)
	            ypred = model(x)
	            return logitcrossentropy(ypred .* trainmask, y .* trainmask, dims=2)
	        end
	        function lossval(x, y)
	            ypred = model(x)
	            return logitcrossentropy(ypred .* valmask, y .* valmask, dims=2)
	        end

			function callback()
			end
			println("Training start.")
	        @epochs args.epochs customtrain!(losstrain, lossval, params(model), feats, labels, optimiser; cb=() -> callback())

			model = cpu(model)
	        feats = cpu(feats)
	        labels = cpu(labels)
			trainmask = cpu(trainmask)
	        valmask = cpu(valmask)

			trainacc = accuracy(feats, labels, model, trainmask)
	        push!(trainaccs, trainacc)
	        valacc = accuracy(feats, labels, model, valmask)
	        push!(valaccs, valacc)
	        testacc = accuracy(feats, labels, model, testmask)
	        push!(testaccs, testacc)


			valpreds = model(feats) .* valmask
			open(valpredf, "w") do fo
				for i in 1:size(valpreds)[1]
					println(fo, valpreds[i, :])
				end
			end

			preds = model(feats) .* testmask
			open(predf, "w") do fo
				for i in 1:size(preds)[1]
					println(fo, preds[i, :])
				end
			end

	        open(resultsfname, "w") do fo
	            println(fo, args)
	            write(fo, "Training accuracy: $(trainacc)\n")
	            write(fo, "Val accuracy: $(valacc)\n")
	            write(fo, "Test accuracy: $(testacc)\n")
	        end

			open(logf, "w") do logfo
				println(logfo, "ran this already")
			end


		end
		resultsavgf = "$resoutdir/$(golabelf[1:end-4])_lr$(args.η)_$(mode).results.avg"
		open(resultsavgf, "w") do fo
			println(fo, "Training accuracy: $(mean(trainaccs)) ± $(std(trainaccs)), $(trainaccs)")
			println(fo, "Val accuracy: $(mean(valaccs)) ± $(std(valaccs)), $(valaccs)")
			println(fo, "Test accuracy: $(mean(testaccs)) ± $(std(testaccs)), $(testaccs)")
		end
		println("Training accuracy: $(mean(trainaccs)) ± $(std(trainaccs))")
		println("Val accuracy: $(mean(valaccs)) ± $(std(valaccs))")
		println("Test accuracy: $(mean(testaccs)) ± $(std(testaccs))")

		if writeaverage

		end
		writeaverage = true


		println("only doing the first GO term for now, comment out the following 'break' to run for all GO terms")
		break;
	end
end

u_args = Args(η=u_lr, epochs=u_epochs)
sign_pipeline(u_args, u_outdir, u_resoutdir, u_godir, u_mode, u_l2n, 2, 0, 1; numprocs=16, seedrange=1:3)
