import tensorflow as tf;
from tensorflow.keras import activations, initializers, regularizers, constraints;
from tensorflow.keras import backend as K;
from tensorflow.keras.layers import Layer;

import numpy as np;
from scipy import sparse;


# # helper functions for preprocessing admats
def degree_power(A, k):
	degrees = np.power(np.array(A.sum(1)), k).flatten();
	degrees[np.isinf(degrees)] = 0.
	D = np.diag(degrees);
	return D;


def normalized_adjacency(A, symmetric=True):
	if symmetric:
		normalized_D = degree_power(A, -0.5)
		output = normalized_D.dot(A).dot(normalized_D);
	else:
		normalized_D = degree_power(A, -1.)
		output = normalized_D.dot(A)
	return output;


# # level 2 part of the NoN-GCN
class l2NoNGCN(Layer):
	def __init__(self,
				 channels,
				 activation=None,
				 use_bias=True,
				 kernel_initializer='glorot_uniform',
				 bias_initializer='zeros',
				 kernel_regularizer=None,
				 bias_regularizer=None,
				 activity_regularizer=None,
				 kernel_constraint=None,
				 bias_constraint=None,
				 **kwargs):

		super().__init__(activity_regularizer=activity_regularizer, **kwargs)
		self.channels = channels
		self.activation = activations.get(activation)
		self.use_bias = use_bias
		self.kernel_initializer = initializers.get(kernel_initializer)
		self.bias_initializer = initializers.get(bias_initializer)
		self.kernel_regularizer = regularizers.get(kernel_regularizer)
		self.bias_regularizer = regularizers.get(bias_regularizer)
		self.kernel_constraint = constraints.get(kernel_constraint)
		self.bias_constraint = constraints.get(bias_constraint)
		self.supports_masking = False

	def build(self, input_shape):
		assert len(input_shape) >= 2
		input_dim = input_shape[0][-1];
		self.kernel = self.add_weight(shape=(input_dim, self.channels),
									  initializer=self.kernel_initializer,
									  name='kernel',
									  regularizer=self.kernel_regularizer,
									  constraint=self.kernel_constraint)

		if self.use_bias:
			self.bias = self.add_weight(shape=(self.channels, ),
				initializer=self.bias_initializer,
				name='bias',
				regularizer=self.bias_regularizer,
				constraint=self.bias_constraint)
		else:
			self.bias = None
		self.built = True;


	def call(self, inputs):
		l2_features = inputs[0];
		l1_features = inputs[1]
		l2_fltr = inputs[2];
		l1_fltr = inputs[3];


		# # aggregate level 1 nodes' features into a summary feature of the level 1 network
		l1_combfeats = tf.math.reduce_mean(l1_features, axis=1);

		# # aggregate level 2 nodes' features with their corresponding level 1 networks' features
		l2_output = tf.tensordot(l2_fltr, l2_features + l1_combfeats, axes=1)
		# # propagate
		l2_output = tf.tensordot(l2_output, self.kernel, axes=1);

		if self.use_bias:
			l2_output = K.bias_add(l2_output, self.bias);
		if self.activation is not None:
			l2_output = self.activation(l2_output);
		return l2_output;

	def compute_output_shape(self, input_shape):
		features_shape = input_shape[0];
		output_shape = features_shape[:-1] + (self.channels, )
		return output_shape;

	def get_config(self):
		config = {
			'channels': self.channels,
			'activation': activations.serialize(self.activation),
			'use_bias': self.use_bias,
			'kernel_initializer': initializers.serialize(self.kernel_initializer),
			'bias_initializer': initializers.serialize(self.bias_initializer),
			'kernel_regularizer': regularizers.serialize(self.kernel_regularizer),
			'bias_regularizer': regularizers.serialize(self.bias_regularizer),
			'kernel_constraint': constraints.serialize(self.kernel_constraint),
			'bias_constraint': constraints.serialize(self.bias_constraint)
		}
		base_config = super().get_config();
		return dict(list(base_config.items()) + list(config.items()))



	@staticmethod
	def preprocess(A):
		symmetric = True;
		fl = A.copy();
		I = np.eye(A.shape[-1], dtype=A.dtype);
		if A.ndim == 3:
			for i in range(A.shape[0]):
				A_tilde = A[i] + I;
				fl[i] = normalized_adjacency(A_tilde, symmetric=symmetric)
		else:
			A_tilde = A + I
			fl = normalized_adjacency(A_tilde, symmetric=symmetric)

		return fl;


class l1NoNGCN(Layer):
	def __init__(self,
				 channels,
				 activation=None,
				 use_bias=True,
				 kernel_initializer='glorot_uniform',
				 bias_initializer='zeros',
				 kernel_regularizer=None,
				 bias_regularizer=None,
				 activity_regularizer=None,
				 kernel_constraint=None,
				 bias_constraint=None,
				 **kwargs):

		super().__init__(activity_regularizer=activity_regularizer, **kwargs)
		self.channels = channels
		self.activation = activations.get(activation)
		self.use_bias = use_bias
		self.kernel_initializer = initializers.get(kernel_initializer)
		self.bias_initializer = initializers.get(bias_initializer)
		self.kernel_regularizer = regularizers.get(kernel_regularizer)
		self.bias_regularizer = regularizers.get(bias_regularizer)
		self.kernel_constraint = constraints.get(kernel_constraint)
		self.bias_constraint = constraints.get(bias_constraint)
		self.supports_masking = False

	def build(self, input_shape):
		assert len(input_shape) >= 2
		input_dim = input_shape[0][-1];
		self.kernel = self.add_weight(shape=(input_dim, self.channels),
									  initializer=self.kernel_initializer,
									  name='kernel',
									  regularizer=self.kernel_regularizer,
									  constraint=self.kernel_constraint)

		if self.use_bias:
			self.bias = self.add_weight(shape=(self.channels, ),
				initializer=self.bias_initializer,
				name='bias',
				regularizer=self.bias_regularizer,
				constraint=self.bias_constraint)
		else:
			self.bias = None
		self.built = True;

	def call(self, inputs):
		l2_features = inputs[0];
		l1_features = inputs[1]
		l2_fltr = inputs[2];
		l1_fltr = inputs[3];


		# # format level 2 nodes' features
		l2_features_repeated = tf.reshape(l2_features, (-1, 1, l2_features.shape[1]))

		# # aggregate level 1 nodes' features with their level 2 nodes' features
		l1_output = tf.matmul(l1_fltr, l1_features + l2_features_repeated);
		l1_output = tf.matmul(l1_output, self.kernel);

		if self.use_bias:
			l1_output = K.bias_add(l1_output, self.bias);
		if self.activation is not None:
			l1_output = self.activation(l1_output);
		return l1_output;

	def compute_output_shape(self, input_shape):
		features_shape = input_shape[0];
		output_shape = features_shape[:-1] + (self.channels, )
		return output_shape;

	def get_config(self):
		config = {
			'channels': self.channels,
			'activation': activations.serialize(self.activation),
			'use_bias': self.use_bias,
			'kernel_initializer': initializers.serialize(self.kernel_initializer),
			'bias_initializer': initializers.serialize(self.bias_initializer),
			'kernel_regularizer': regularizers.serialize(self.kernel_regularizer),
			'bias_regularizer': regularizers.serialize(self.bias_regularizer),
			'kernel_constraint': constraints.serialize(self.kernel_constraint),
			'bias_constraint': constraints.serialize(self.bias_constraint)
		}
		base_config = super().get_config();
		return dict(list(base_config.items()) + list(config.items()))



	@staticmethod
	def preprocess(A):
		#print("preprocing l1")
		symmetric = True;
		fl = A.copy();
		I = np.eye(A.shape[-1], dtype=A.dtype);
		if A.ndim == 3:
			for i in range(A.shape[0]):
				A_tilde = A[i] + I;
				fl[i] = normalized_adjacency(A_tilde, symmetric=symmetric)
		else:
			A_tilde = A + I
			fl = normalized_adjacency(A_tilde, symmetric=symmetric)

		return fl;



# # GCN, adapted from https://github.com/danielegrattarola/spektral
class StandardGCN(Layer):
	def __init__(self,
				 channels,
				 activation=None,
				 use_bias=True,
				 kernel_initializer='glorot_uniform',
				 bias_initializer='zeros',
				 kernel_regularizer=None,
				 bias_regularizer=None,
				 activity_regularizer=None,
				 kernel_constraint=None,
				 bias_constraint=None,
				 **kwargs):

		super().__init__(activity_regularizer=activity_regularizer, **kwargs)
		self.channels = channels
		self.activation = activations.get(activation)
		self.use_bias = use_bias
		self.kernel_initializer = initializers.get(kernel_initializer)
		self.bias_initializer = initializers.get(bias_initializer)
		self.kernel_regularizer = regularizers.get(kernel_regularizer)
		self.bias_regularizer = regularizers.get(bias_regularizer)
		self.kernel_constraint = constraints.get(kernel_constraint)
		self.bias_constraint = constraints.get(bias_constraint)
		self.supports_masking = False

	def build(self, input_shape):
		assert len(input_shape) >= 2
		input_dim = input_shape[0][-1];
		self.kernel = self.add_weight(shape=(input_dim, self.channels),
									  initializer=self.kernel_initializer,
									  name='kernel',
									  regularizer=self.kernel_regularizer,
									  constraint=self.kernel_constraint)

		if self.use_bias:
			self.bias = self.add_weight(shape=(self.channels, ),
				initializer=self.bias_initializer,
				name='bias',
				regularizer=self.bias_regularizer,
				constraint=self.bias_constraint)
		else:
			self.bias = None
		self.built = True;

	def call(self, inputs):
		features = inputs[0];
		test = inputs[1];
		fltr = inputs[2];

		output = tf.tensordot(features, self.kernel, axes=1);
		output = tf.tensordot(fltr, output, axes=1);

		if self.use_bias:
			output = K.bias_add(output, self.bias);
		if self.activation is not None:
			output = self.activation(output);
		return output, test;

	def compute_output_shape(self, input_shape):
		features_shape = input_shape[0];
		output_shape = features_shape[:-1] + (self.channels, )
		return output_shape;

	def get_config(self):
		config = {
			'channels': self.channels,
			'activation': activations.serialize(self.activation),
			'use_bias': self.use_bias,
			'kernel_initializer': initializers.serialize(self.kernel_initializer),
			'bias_initializer': initializers.serialize(self.bias_initializer),
			'kernel_regularizer': regularizers.serialize(self.kernel_regularizer),
			'bias_regularizer': regularizers.serialize(self.bias_regularizer),
			'kernel_constraint': constraints.serialize(self.kernel_constraint),
			'bias_constraint': constraints.serialize(self.bias_constraint)
		}
		base_config = super().get_config();
		return dict(list(base_config.items()) + list(config.items()))



	@staticmethod
	def preprocess(A):
		symmetric = True;
		fl = A.copy();
		I = np.eye(A.shape[-1], dtype=A.dtype);
		if A.ndim == 3:
			for i in range(A.shape[0]):
				A_tilde = A[i] + I;
				fl[i] = normalized_adjacency(A_tilde, symmetric=symmetric)
		else:
			A_tilde = A + I
			fl = normalized_adjacency(A_tilde, symmetric=symmetric)

		return fl;

   