using ArgParse
using LightGraphs
using Flux.Data: DataLoader
using Flux: onecold, logitcrossentropy, @epochs, train!, Dense, Chain, ADAM, params, update!, gpu, cpu, softmax
import Zygote: Params, gradient
println("loaded Flux")
# println("not loading Flux until necessary")
# push!(LOAD_PATH, "../FluxGNNs/src/")
push!(LOAD_PATH, "../")
using FluxGNNs
using LinearAlgebra
using SparseArrays
using Statistics
using StatsBase
using Parameters: @with_kw
using SyntheticNoNs
using YAML
using CUDA
if has_cuda()
    @info "CUDA is on"
    CUDA.allowscalar(false)
end
push!(LOAD_PATH, "./scripts/")
using ProteinNoNLoader
include("./utils.jl")
println("finished loading packages")


function parsecli()
	apsettings = ArgParseSettings()
	@add_arg_table apsettings begin
		# "--configf", "-c"
			# default = "./configs/test.yml"
		"--method", "-m"
			default = "sign"
			help = "SIGN-based approach you want to run"
		"--epochs", "-e"
			arg_type = Int
			default = 100
			help = "number of epochs to train for"
		"--config", "-c"
			default = ""
			help = "path to .yml file containing NoN information"
	end
	return parse_args(ARGS, apsettings)
end
parsedargs = parsecli()
println(parsedargs)
u_method = parsedargs["method"]
u_epochs = parsedargs["epochs"]
u_configf = parsedargs["config"]

# # struct containing arguments
@with_kw mutable struct Args
    η::Float64 = 0.01
    batchsize::Int = 1024
    epochs::Int = 10
	channels::Int = 256
    device::Function = gpu
end
Base.show(io::IO, args::Args) = print(io, "Learning rate: $(args.η)\nBatch size: $(args.batchsize)\nEpochs: $(args.epochs)\nDevice: $(args.device)")


# # preprocess adjacency matrix, as done in Kipf and Welling (2016).
function preprocess(adjmat::AbstractArray)
    adjmathat = adjmat
    degrees = sum(adjmathat, dims=2)
    degrees = vec(degrees) .^ -0.5
    normD = spdiagm(0 => degrees)
    return normD * adjmathat * normD
end

# # calculate accuracy from input features
function accuracy(x, y, model, mask)
    ypred = model(x)
	return accuracy(ypred, y, mask)
end

# # calculate accuracy from predictions
function accuracy(ypred, y, mask)
	acc = 0
	for i in 1:size(y)[1]
		# # only consider entities that are for the given training/val/testing set
        if mask[i]
            acc += sum(onecold(ypred[i, :]) == onecold(y[i, :]))
        end
    end
    return acc/sum(mask)
end

# # training function for Flux. See https://fluxml.ai/Flux.jl/stable/training/training/ for more information.
function customtrain!(losstrain, lossval, ps, feats, labels, opt; cb= () -> ())
    t0 = time()
    ps = Params(ps)

    try
        gs = gradient(ps) do
            losstrain(feats, labels)
        end
        update!(opt, ps, gs)
        cb()
    catch ex
        rethrow(ex)
    end
    println("Training loss: $(round(losstrain(feats, labels); digits=4))")
    println("Val loss: $(round(lossval(feats, labels); digits=4))")
    println("Epoch time: $(time() - t0)")
end


# # get SIGN's triangle motif adjacency matrix
# # adjlistf is the adjacency list of the network. It is needed as input into the triangle motif computation program.
# # adjmat is the adjacency matrix, as a sparse matrix
function get_tri_adjmat(adjlistf, adjmat)

	# # format the adjacency matrix into the .mtx format that PGD prefers
	tempf = "$(adjlistf[1:end-4]).mtx"
	edgecount = 0
	nodeset = Set()
	open(adjlistf, "r") do fi
		for line in eachline(fi)
			larr = split(line)
			node1 = larr[1]
			node2 = larr[2]
			push!(nodeset, node1)
			push!(nodeset, node2)
			edgecount += 1
		end
	end

	nnodes = length(nodeset)
	open(tempf, "w") do fo
		println(fo, nnodes, "\t", nnodes, "\t", edgecount)
		write(fo, read(adjlistf, String))
	end

	# # run PGD to compute triangle motifs, create a temporary file for the output
	outf = "$tempf.micro"
	open("junk.txt", "w") do fo
		write(fo, read(`./scripts/pgd -f $tempf --micro $outf`))
	end


	triadjmat = deepcopy(adjmat)
	# # read the output of the temporary file for the motif counts, and create the triangle motif adjacency matrix
	open(outf, "r") do fi
		readline(fi)
		for line in eachline(fi)
			larr = parse.(Int, split(line, ","))
			n1 = larr[1]
			n2 = larr[2]
			numtris = larr[3]
			triadjmat[n1, n2] = numtris * triadjmat[n1, n2]
			triadjmat[n2, n1] = numtris * triadjmat[n2, n1]
		end
	end
	run(`rm $tempf`)

	return triadjmat
end


# # nonparams is a struct containing the information about the NoN
# # args is a struct containing the parameters of the classifier
# # mode is the approach, i.e., sign
# # ppow is the expontent of the adjacency matrix, spow is the exponent of the pagerank adjacency matrix, tpow is the exponent of the triangle motif adjacency matrix, as defined in SIGN
# # seedrange corresponds to which random seeds to use, which determine the training/val/testing split
# # numprocs is the number of processes to use
function sign_pipeline(nonparams, args, mode, ppow, spow, tpow; seedrange=1:1, numprocs=1)


    nondir = nonparams["nondir"]
	l2n = nonparams["l2n"]
    labels = loadlabels("$nondir/labels.onehot.txt")

    # # get adjacency matrix and preprocess it
	G = loadedgelist("$nondir/l2/l2.txt")
	adj = adjacency_matrix(G)
	adjhat = preprocess(adj + sparse(I, size(adj)))
	# # concatente the adjmat with adjmat^2, adjmat^3, etc.
	featholder = []
	for i in 1:ppow
		push!(featholder, adjhat ^ i)
	end

	# # triangle motif adjmat. SIGN does not use the pagerank adjmat for the biological network it tests, so we don't include that here
	triadjmat = get_tri_adjmat("$nondir/l2/l2.txt", adj)
	triadjmat = preprocess(triadjmat + sparse(I, size(triadjmat)))
	for i in 1:tpow
		push!(featholder, triadjmat ^ i)
	end

	l1dir = "$nondir/l1/"

    trainaccs = []
    valaccs = []
    testaccs = []

    # # reshape data into format that Flux uses
	feats = reduce(hcat, featholder)
	labels = reduce(hcat, labels)'

	# # for each training/val/testing split
    for seed in seedrange

    	# # name of output results file
		resultsfname = "$nondir/lr$(args.η)_$(mode)_seed$(seed).results"
		# # if sign was already run for this random seed, just load its results but don't rerun
		if isfile(resultsfname)
			println("$resultsfname already exists")
			open(resultsfname, "r") do fi
				readline(fi)
				readline(fi)
				readline(fi)
				readline(fi)
				tracc = parse(Float64, split(readline(fi), ": ")[2])
				push!(trainaccs, tracc)
				valacc = parse(Float64, split(readline(fi), ": ")[2])
				push!(valaccs, valacc)
				teacc = parse(Float64, split(readline(fi), ": ")[2])
				push!(testaccs, teacc)
			end
			continue
		end

		# # julia starts from index 1, python starts from index 0
		masknum = seed-1
		# # load the same masks used by the graphlet features, so that the training/val/testing splits are fairly comparable
		trmaskf = "$nondir/_lr0.1_l2gdv_seed$(masknum)_norm_trmask.txt"
		trainmask = loadmask(trmaskf, nnodes=l2n)
		valmaskf = "$nondir/_lr0.1_l2gdv_seed$(masknum)_norm_valmask.txt"
		valmask = loadmask(valmaskf, nnodes=l2n)
		temaskf = "$nondir/_lr0.1_l2gdv_seed$(masknum)_norm_temask.txt"
		testmask = loadmask(temaskf, nnodes=l2n)
		# # quick check that the masks are disjoint, totalling up to the number of entities
		println(sum(trainmask .⊻ valmask .⊻ testmask))
        
		# # information
        println("featsize $(size(feats))")
		println("feat type $(typeof(feats))")
        println("labelsize $(size(labels))")
		println("label type $(typeof(labels))")
		println("masksizes $(size(trainmask)), $(size(valmask))")
		println("mask type $(typeof(trainmask)), $(typeof(valmask))")

        featsize = size(feats)[2]
        numlabels = size(labels)[2]


        # # logistic regression classifier in Flux
		lrlayer = DenseEnd(featsize, numlabels)
		model = Chain(lrlayer, x->softmax(x, dims=2))

        optimiser = ADAM(args.η)

        # # if gpu is available, move data to gpu for computation
        model = args.device(model)
        feats = args.device(feats)
        labels = args.device(labels)

        trainmask = args.device(trainmask)
        valmask = args.device(valmask)


        # # loss functions using categorical crossentropy. Use masks instead of disjoint datasets.
        function losstrain(x, y)
            ypred = model(x)
            return logitcrossentropy(ypred .* trainmask, y .* trainmask, dims=2) # works if onehot is along columns. Use dims=2 if onehot is along rows
        end
        function lossval(x, y)
            ypred = model(x)
            return logitcrossentropy(ypred .* valmask, y .* valmask, dims=2)
        end

        println("Training start.")
        @epochs args.epochs customtrain!(losstrain, lossval, params(model), feats, labels, optimiser)
        
        # # move data back to CPU for accuracy calculations
        model = cpu(model)
        feats = cpu(feats)
        labels = cpu(labels)

        trainmask = cpu(trainmask)
        valmask = cpu(valmask)

        trainacc = accuracy(feats, labels, model, trainmask)
        push!(trainaccs, trainacc)
        valacc = accuracy(feats, labels, model, valmask)
        push!(valaccs, valacc)
        testacc = accuracy(feats, labels, model, testmask)
        push!(testaccs, testacc)


        # # write renults out to fcile
		println(resultsfname)
        open(resultsfname, "w") do fo
            println(fo, args)
            write(fo, "Training accuracy: $(trainacc)\n")
            write(fo, "Val accuracy: $(valacc)\n")
            write(fo, "Test accuracy: $(testacc)\n")
        end


        # # save SIGN features for reference
		if mode == "sign"
			outputfeats = feats
			open("$nondir/l2/sign_feats_seed$(seed).txt", "w") do fo
				for i in 1:size(outputfeats)[1]
					println(fo, join(outputfeats[i, :], "\t"))
				end
			end
		end

    end
    # # average over all random samplings
    println("Training accuracy: $(mean(trainaccs)) ± $(std(trainaccs))")
    println("Val accuracy: $(mean(valaccs)) ± $(std(valaccs))")
    println("Test accuracy: $(mean(testaccs)) ± $(std(testaccs))")
end


println(u_configf)
g_nonparams = readyml(u_configf)
g_args = Args(η=0.01, batchsize=g_nonparams["l2n"], epochs=u_epochs)
sign_pipeline(g_nonparams, g_args, u_method, 2, 0, 1; seedrange=1:3, numprocs=1)

